@@ -31,14 +31,16 @@
 #include <glib.h>
 
 #include <ofono/log.h>
-#include "driver.h"
+#include <ofono/modem.h>
+#include <ofono/sms.h>
 #include "smsutil.h"
 #include "util.h"
+#include "vendor.h"
 
 #include "gatchat.h"
 #include "gatresult.h"
 
-#include "at.h"
+#include "atmodem.h"
 
 static const char *csca_prefix[] = { "+CSCA:", NULL };
 static const char *csms_prefix[] = { "+CSMS:", NULL };
@@ -46,10 +48,12 @@ static const char *cmgf_prefix[] = { "+CMGF:", NULL };
 static const char *cpms_prefix[] = { "+CPMS:", NULL };
 static const char *cnmi_prefix[] = { "+CNMI:", NULL };
 static const char *cmgs_prefix[] = { "+CMGS:", NULL };
+static const char *cmgl_prefix[] = { "+CMGL:", NULL };
 static const char *none_prefix[] = { NULL };
 
 static gboolean set_cmgf(gpointer user_data);
 static gboolean set_cpms(gpointer user_data);
+static void at_cmgl_set_cpms(struct ofono_sms *sms, int store);
 
 #define MAX_CMGF_RETRIES 10
 #define MAX_CPMS_RETRIES 10
@@ -71,31 +75,20 @@ struct sms_data {
 	gboolean cnma_enabled;
 	char *cnma_ack_pdu;
 	int cnma_ack_pdu_len;
+	GAtChat *chat;
+	unsigned int vendor;
 };
 
 struct cpms_request {
-	struct ofono_modem *modem;
+	struct ofono_sms *sms;
 	int store;
 	int index;
 };
 
-static struct sms_data *sms_create()
-{
-	return g_try_new0(struct sms_data, 1);
-}
-
-static void sms_destroy(struct sms_data *data)
-{
-	if (data->cnma_ack_pdu)
-		g_free(data->cnma_ack_pdu);
-
-	g_free(data);
-}
-
 static void at_csca_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
-	ofono_generic_cb_t cb = cbd->cb;
+	ofono_sms_sca_set_cb_t cb = cbd->cb;
 	struct ofono_error error;
 
 	dump_response("csca_set_cb", ok, result);
@@ -104,12 +97,12 @@ static void at_csca_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	cb(&error, cbd->data);
 }
 
-static void at_csca_set(struct ofono_modem *modem,
+static void at_csca_set(struct ofono_sms *sms,
 			const struct ofono_phone_number *sca,
-			ofono_generic_cb_t cb, void *data)
+			ofono_sms_sca_set_cb_t cb, void *user_data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
-	struct cb_data *cbd = cb_data_new(modem, cb, data);
+	struct sms_data *data = ofono_sms_get_data(sms);
+	struct cb_data *cbd = cb_data_new(cb, user_data);
 	char buf[64];
 
 	if (!cbd)
@@ -117,7 +110,7 @@ static void at_csca_set(struct ofono_modem *modem,
 
 	sprintf(buf, "AT+CSCA=\"%s\",%d", sca->number, sca->type);
 
-	if (g_at_chat_send(at->parser, buf, csca_prefix,
+	if (g_at_chat_send(data->chat, buf, csca_prefix,
 				at_csca_set_cb, cbd, g_free) > 0)
 		return;
 
@@ -125,17 +118,14 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, user_data);
 }
 
 static void at_csca_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
 	GAtResultIter iter;
-	ofono_sca_query_cb_t cb = cbd->cb;
+	ofono_sms_sca_query_cb_t cb = cbd->cb;
 	struct ofono_error error;
 	struct ofono_phone_number sca;
 	const char *number;
@@ -174,22 +164,19 @@ static void at_csca_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	return;
 
 err:
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, NULL, cbd->data);
-	}
+	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
 }
 
-static void at_csca_query(struct ofono_modem *modem, ofono_sca_query_cb_t cb,
-					void *data)
+static void at_csca_query(struct ofono_sms *sms, ofono_sms_sca_query_cb_t cb,
+					void *user_data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
-	struct cb_data *cbd = cb_data_new(modem, cb, data);
+	struct sms_data *data = ofono_sms_get_data(sms);
+	struct cb_data *cbd = cb_data_new(cb, user_data);
 
 	if (!cbd)
 		goto error;
 
-	if (g_at_chat_send(at->parser, "AT+CSCA?", csca_prefix,
+	if (g_at_chat_send(data->chat, "AT+CSCA?", csca_prefix,
 				at_csca_query_cb, cbd, g_free) > 0)
 		return;
 
@@ -197,10 +184,7 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, NULL, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, NULL, user_data);
 }
 
 static void at_cmgs_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -233,32 +217,31 @@ static void at_cmgs_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	return;
 
 err:
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, -1, cbd->data);
-	}
+	CALLBACK_WITH_FAILURE(cb, -1, cbd->data);
 }
 
-static void at_cmgs(struct ofono_modem *modem, unsigned char *pdu, int pdu_len,
-			int tpdu_len, gboolean mms, ofono_sms_submit_cb_t cb,
-			void *data)
+static void at_cmgs(struct ofono_sms *sms, unsigned char *pdu, int pdu_len,
+			int tpdu_len, int mms, ofono_sms_submit_cb_t cb,
+			void *user_data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
-	struct cb_data *cbd = cb_data_new(modem, cb, data);
+	struct sms_data *data = ofono_sms_get_data(sms);
+	struct cb_data *cbd = cb_data_new(cb, user_data);
 	char buf[512];
 	int len;
 
 	if (!cbd)
 		goto error;
 
-	if (mms)
-		g_at_chat_send(at->parser, "AT+CMMS=1", none_prefix,
+	if (mms) {
+		sprintf(buf, "AT+CMMS=%d", mms);
+		g_at_chat_send(data->chat, buf, none_prefix,
 				NULL, NULL, NULL);
+	}
 
 	len = sprintf(buf, "AT+CMGS=%d\r", tpdu_len);
 	encode_hex_own_buf(pdu, pdu_len, 0, buf+len);
 
-	if (g_at_chat_send(at->parser, buf, cmgs_prefix,
+	if (g_at_chat_send(data->chat, buf, cmgs_prefix,
 				at_cmgs_cb, cbd, g_free) > 0)
 		return;
 
@@ -266,18 +249,9 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, -1, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, -1, user_data);
 }
 
-static struct ofono_sms_ops ops = {
-	.sca_query	= at_csca_query,
-	.sca_set	= at_csca_set,
-	.submit		= at_cmgs,
-};
-
 static void at_cnma_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	if (!ok)
@@ -306,25 +280,10 @@ static gboolean at_parse_pdu_common(GAtResult *result, const char *prefix,
 	return TRUE;
 }
 
-static void at_cbm_notify(GAtResult *result, gpointer user_data)
-{
-	int pdulen;
-	const char *pdu;
-
-	dump_response("at_cbm_notify", TRUE, result);
-
-	if (!at_parse_pdu_common(result, "+CBM:", &pdu, &pdulen)) {
-		ofono_error("Unable to parse CBM notification");
-		return;
-	}
-
-	ofono_debug("Got new Cell Broadcast via CBM: %s, %d", pdu, pdulen);
-}
-
 static void at_cds_notify(GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	int pdulen;
 	const char *pdu;
 	char buf[256];
@@ -339,23 +298,23 @@ static void at_cds_notify(GAtResult *result, gpointer user_data)
 	ofono_debug("Got new Status-Report PDU via CDS: %s, %d", pdu, pdulen);
 
 	/* We must acknowledge the PDU using CNMA */
-	if (at->sms->cnma_ack_pdu)
-		sprintf(buf, "AT+CNMA=1,%d\r%s", at->sms->cnma_ack_pdu_len,
-				at->sms->cnma_ack_pdu);
+	if (data->cnma_ack_pdu)
+		sprintf(buf, "AT+CNMA=1,%d\r%s", data->cnma_ack_pdu_len,
+				data->cnma_ack_pdu);
 	else
 		sprintf(buf, "AT+CNMA=0"); /* Should be a safe fallback */
 
-	g_at_chat_send(at->parser, buf, none_prefix, at_cnma_cb, NULL, NULL);
+	g_at_chat_send(data->chat, buf, none_prefix, at_cnma_cb, NULL, NULL);
 }
 
 static void at_cmt_notify(GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	const char *hexpdu;
 	long pdu_len;
 	int tpdu_len;
-	unsigned char pdu[164];
+	unsigned char pdu[176];
 	char buf[256];
 
 	dump_response("at_cmt_notify", TRUE, result);
@@ -365,27 +324,32 @@ static void at_cmt_notify(GAtResult *result, gpointer user_data)
 		return;
 	}
 
+	if (strlen(hexpdu) > sizeof(pdu) * 2) {
+		ofono_error("Bad PDU length in CMT notification");
+		return;
+	}
+
 	ofono_debug("Got new SMS Deliver PDU via CMT: %s, %d", hexpdu, tpdu_len);
 
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
-	ofono_sms_deliver_notify(modem, pdu, pdu_len, tpdu_len);
+	ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
 
 	/* We must acknowledge the PDU using CNMA */
-	if (at->sms->cnma_ack_pdu)
-		sprintf(buf, "AT+CNMA=1,%d\r%s", at->sms->cnma_ack_pdu_len,
-				at->sms->cnma_ack_pdu);
+	if (data->cnma_ack_pdu)
+		sprintf(buf, "AT+CNMA=1,%d\r%s", data->cnma_ack_pdu_len,
+				data->cnma_ack_pdu);
 	else
 		sprintf(buf, "AT+CNMA=0"); /* Should be a safe fallback */
 
-	g_at_chat_send(at->parser, buf, none_prefix, at_cnma_cb, NULL, NULL);
+	g_at_chat_send(data->chat, buf, none_prefix, at_cnma_cb, NULL, NULL);
 }
 
 static void at_cmgr_notify(GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
+	struct ofono_sms *sms = user_data;
 	GAtResultIter iter;
 	const char *hexpdu;
-	unsigned char pdu[164];
+	unsigned char pdu[176];
 	long pdu_len;
 	int tpdu_len;
 
@@ -407,10 +371,13 @@ static void at_cmgr_notify(GAtResult *result, gpointer user_data)
 
 	hexpdu = g_at_result_pdu(result);
 
+	if (strlen(hexpdu) > sizeof(pdu) * 2)
+		goto err;
+
 	ofono_debug("Got PDU: %s, with len: %d", hexpdu, tpdu_len);
 
 	decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
-	ofono_sms_deliver_notify(modem, pdu, pdu_len, tpdu_len);
+	ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
 	return;
 
 err:
@@ -432,8 +399,8 @@ static void at_cmgd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 static void at_cmti_cpms_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cpms_request *req = user_data;
-	struct ofono_modem *modem = req->modem;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = req->sms;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	char buf[128];
 
 	if (!ok) {
@@ -441,20 +408,20 @@ static void at_cmti_cpms_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	at->sms->store = req->store;
+	data->store = req->store;
 
 	sprintf(buf, "AT+CMGR=%d", req->index);
-	g_at_chat_send(at->parser, buf, none_prefix, at_cmgr_cb, modem, NULL);
+	g_at_chat_send(data->chat, buf, none_prefix, at_cmgr_cb, NULL, NULL);
 
 	/* We don't buffer SMS on the SIM/ME, send along a CMGD as well */
 	sprintf(buf, "AT+CMGD=%d", req->index);
-	g_at_chat_send(at->parser, buf, none_prefix, at_cmgd_cb, NULL, NULL);
+	g_at_chat_send(data->chat, buf, none_prefix, at_cmgd_cb, NULL, NULL);
 }
 
 static void at_cmti_notify(GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	const char *strstore;
 	int store;
 	GAtResultIter iter;
@@ -482,27 +449,27 @@ static void at_cmti_notify(GAtResult *result, gpointer user_data)
 
 	ofono_debug("Got a CMTI indication at %s, index: %d", strstore, index);
 
-	if (store == at->sms->store) {
+	if (store == data->store) {
 		struct cpms_request req;
 
-		req.modem = modem;
+		req.sms = sms;
 		req.store = store;
 		req.index = index;
 
 		at_cmti_cpms_cb(TRUE, NULL, &req);
 	} else {
 		char buf[128];
-		const char *incoming = storages[at->sms->incoming];
+		const char *incoming = storages[data->incoming];
 		struct cpms_request *req = g_new(struct cpms_request, 1);
 
-		req->modem = modem;
+		req->sms = sms;
 		req->store = store;
 		req->index = index;
 
 		sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"",
 			strstore, strstore, incoming);
 
-		g_at_chat_send(at->parser, buf, cpms_prefix, at_cmti_cpms_cb,
+		g_at_chat_send(data->chat, buf, cpms_prefix, at_cmti_cpms_cb,
 				req, g_free);
 	}
 
@@ -512,46 +479,166 @@ err:
 	ofono_error("Unable to parse CMTI notification");
 }
 
-static void at_sms_initialized(struct ofono_modem *modem)
+static void at_cmgl_done(struct ofono_sms *sms)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct sms_data *data = ofono_sms_get_data(sms);
 
-	g_at_chat_register(at->parser, "+CMTI:", at_cmti_notify, FALSE,
-				modem, NULL);
-	g_at_chat_register(at->parser, "+CMT:", at_cmt_notify, TRUE,
-				modem, NULL);
-	g_at_chat_register(at->parser, "+CDS:", at_cds_notify, TRUE,
-				modem, NULL);
-	g_at_chat_register(at->parser, "+CBM:", at_cbm_notify, TRUE,
-				modem, NULL);
-
-	/* We treat CMGR just like a notification */
-	g_at_chat_register(at->parser, "+CMGR:", at_cmgr_notify, TRUE,
-				modem, NULL);
-
-	ofono_sms_manager_register(modem, &ops);
+	if (data->incoming == MT_STORE && data->store == ME_STORE)
+		at_cmgl_set_cpms(sms, SM_STORE);
 }
 
-static void at_sms_not_supported(struct ofono_modem *modem)
+static void at_cmgl_notify(GAtResult *result, gpointer user_data)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
+	GAtResultIter iter;
+	const char *hexpdu;
+	unsigned char pdu[176];
+	long pdu_len;
+	int tpdu_len;
+	int index;
+	int status;
+	char buf[16];
 
+	dump_response("at_cmgl_notify", TRUE, result);
+
+	g_at_result_iter_init(&iter, result);
+
+	while (g_at_result_iter_next(&iter, "+CMGL:")) {
+		if (!g_at_result_iter_next_number(&iter, &index))
+			goto err;
+
+		if (!g_at_result_iter_next_number(&iter, &status))
+			goto err;
+
+		if (!g_at_result_iter_skip_next(&iter))
+			goto err;
+
+		if (!g_at_result_iter_next_number(&iter, &tpdu_len))
+			goto err;
+
+		/* Only MT messages */
+		if (status != 0 && status != 1)
+			continue;
+
+		hexpdu = g_at_result_pdu(result);
+
+		ofono_debug("Found an old SMS PDU: %s, with len: %d",
+				hexpdu, tpdu_len);
+
+		if (strlen(hexpdu) > sizeof(pdu) * 2)
+			continue;
+
+		decode_hex_own_buf(hexpdu, -1, &pdu_len, 0, pdu);
+		ofono_sms_deliver_notify(sms, pdu, pdu_len, tpdu_len);
+
+		/* We don't buffer SMS on the SIM/ME, send along a CMGD */
+		sprintf(buf, "AT+CMGD=%d", index);
+		g_at_chat_send(data->chat, buf, none_prefix,
+				at_cmgd_cb, NULL, NULL);
+	}
+	return;
+
+err:
+	ofono_error("Unable to parse CMGL response");
+}
+
+static void at_cmgl_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct ofono_sms *sms = user_data;
+
+	if (!ok)
+		ofono_error("Initial listing SMS storage failed!");
+
+	at_cmgl_done(sms);
+}
+
+static void at_cmgl_cpms_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cpms_request *req = user_data;
+	struct ofono_sms *sms = req->sms;
+	struct sms_data *data = ofono_sms_get_data(sms);
+
+	if (!ok) {
+		ofono_error("Initial CPMS request failed");
+		at_cmgl_done(sms);
+		return;
+	}
+
+	data->store = req->store;
+
+	g_at_chat_send_pdu_listing(data->chat, "AT+CMGL=4", cmgl_prefix,
+					at_cmgl_notify, at_cmgl_cb, sms, NULL);
+}
+
+static void at_cmgl_set_cpms(struct ofono_sms *sms, int store)
+{
+	struct sms_data *data = ofono_sms_get_data(sms);
+
+	if (store == data->store) {
+		struct cpms_request req;
+
+		req.sms = sms;
+		req.store = store;
+
+		at_cmgl_cpms_cb(TRUE, NULL, &req);
+	} else {
+		char buf[128];
+		const char *readwrite = storages[store];
+		const char *incoming = storages[data->incoming];
+		struct cpms_request *req = g_new(struct cpms_request, 1);
+
+		req->sms = sms;
+		req->store = store;
+
+		sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"",
+			readwrite, readwrite, incoming);
+
+		g_at_chat_send(data->chat, buf, cpms_prefix, at_cmgl_cpms_cb,
+				req, g_free);
+	}
+}
+
+static void at_sms_initialized(struct ofono_sms *sms)
+{
+	struct sms_data *data = ofono_sms_get_data(sms);
+
+	g_at_chat_register(data->chat, "+CMTI:", at_cmti_notify, FALSE,
+				sms, NULL);
+	g_at_chat_register(data->chat, "+CMT:", at_cmt_notify, TRUE,
+				sms, NULL);
+	g_at_chat_register(data->chat, "+CDS:", at_cds_notify, TRUE,
+				sms, NULL);
+
+	/* We treat CMGR just like a notification */
+	g_at_chat_register(data->chat, "+CMGR:", at_cmgr_notify, TRUE,
+				sms, NULL);
+
+	/* Inspect and free the incoming SMS storage */
+	if (data->incoming == MT_STORE)
+		at_cmgl_set_cpms(sms, ME_STORE);
+	else
+		at_cmgl_set_cpms(sms, data->incoming);
+
+	ofono_sms_register(sms);
+}
+
+static void at_sms_not_supported(struct ofono_sms *sms)
+{
 	ofono_error("SMS not supported by this modem.  If this is in error"
 			" please submit patches to support this hardware");
-	if (at->sms) {
-		sms_destroy(at->sms);
-		at->sms = NULL;
-	}
+
+	ofono_sms_remove(sms);
 }
 
 static void at_cnmi_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
+	struct ofono_sms *sms = user_data;
 
 	if (!ok)
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 
-	at_sms_initialized(modem);
+	at_sms_initialized(sms);
 }
 
 static inline char wanted_cnmi(int supported, const char *pref)
@@ -588,17 +675,25 @@ static inline gboolean append_cnmi_element(char *buf, int *len, int cap,
 }
 
 static gboolean build_cnmi_string(char *buf, int *cnmi_opts,
-					gboolean cnma_enabled)
+					struct sms_data *data)
 {
+	const char *mode;
 	int len = sprintf(buf, "AT+CNMI=");
 
-	/* Mode doesn't matter, but sounds like 2 is the sanest option */
-	if (!append_cnmi_element(buf, &len, cnmi_opts[0], "2310", FALSE))
+	if (data->vendor == OFONO_VENDOR_HTC_G1)
+		/* The G1 advertises support for mode 2, but returns an error
+		 * if we attempt to actually use it. */
+		mode = "1";
+	else
+		/* Sounds like 2 is the sanest mode */
+		mode = "2310";
+
+	if (!append_cnmi_element(buf, &len, cnmi_opts[0], mode, FALSE))
 		return FALSE;
 
 	/* Prefer to deliver SMS via +CMT if CNMA is supported */
 	if (!append_cnmi_element(buf, &len, cnmi_opts[1],
-					cnma_enabled ? "21" : "1", FALSE))
+					data->cnma_enabled ? "21" : "1", FALSE))
 		return FALSE;
 
 	/* Always deliver CB via +CBM, otherwise don't deliver at all */
@@ -648,8 +743,8 @@ err:
 
 static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	GAtResultIter iter;
 	int cnmi_opts[5]; /* See 27.005 Section 3.4.1 */
 	int opt;
@@ -684,101 +779,101 @@ static void at_cnmi_query_cb(gboolean ok, GAtResult *result, gpointer user_data)
 			goto out;
 	}
 
-	if (build_cnmi_string(buf, cnmi_opts, at->sms->cnma_enabled))
+	if (build_cnmi_string(buf, cnmi_opts, data))
 		supported = TRUE;
 
-	if (at->sms->cnma_enabled)
-		construct_ack_pdu(at->sms);
+	if (data->cnma_enabled)
+		construct_ack_pdu(data);
 
 out:
 	if (!supported)
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 
-	g_at_chat_send(at->parser, buf, cnmi_prefix,
-			at_cnmi_set_cb, modem, NULL);
+	g_at_chat_send(data->chat, buf, cnmi_prefix,
+			at_cnmi_set_cb, sms, NULL);
 }
 
-static void at_query_cnmi(struct ofono_modem *modem)
+static void at_query_cnmi(struct ofono_sms *sms)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct sms_data *data = ofono_sms_get_data(sms);
 
-	g_at_chat_send(at->parser, "AT+CNMI=?", cnmi_prefix,
-			at_cnmi_query_cb, modem, NULL);
+	g_at_chat_send(data->chat, "AT+CNMI=?", cnmi_prefix,
+			at_cnmi_query_cb, sms, NULL);
 }
 
 static void at_cpms_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 
 	dump_response("at_cpms_set_cb", ok, result);
 
 	if (ok)
-		return at_query_cnmi(modem);
+		return at_query_cnmi(sms);
 
-	at->sms->retries += 1;
+	data->retries += 1;
 
-	if (at->sms->retries == MAX_CPMS_RETRIES) {
+	if (data->retries == MAX_CPMS_RETRIES) {
 		ofono_error("Unable to set preferred storage");
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 	}
 
-	g_timeout_add_seconds(1, set_cpms, modem);
+	g_timeout_add_seconds(1, set_cpms, sms);
 }
 
 static gboolean set_cpms(gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
-	const char *store = storages[at->sms->store];
-	const char *incoming = storages[at->sms->incoming];
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
+	const char *store = storages[data->store];
+	const char *incoming = storages[data->incoming];
 	char buf[128];
 
 	sprintf(buf, "AT+CPMS=\"%s\",\"%s\",\"%s\"", store, store, incoming);
 
-	g_at_chat_send(at->parser, buf, cpms_prefix,
-			at_cpms_set_cb, modem, NULL);
+	g_at_chat_send(data->chat, buf, cpms_prefix,
+			at_cpms_set_cb, sms, NULL);
 	return FALSE;
 }
 
 static void at_cmgf_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 
 	dump_response("at_cmgf_set_cb", ok, result);
 
 	if (ok) {
-		at->sms->retries = 0;
-		set_cpms(modem);
+		data->retries = 0;
+		set_cpms(sms);
 		return;
 	}
 
-	at->sms->retries += 1;
+	data->retries += 1;
 
-	if (at->sms->retries == MAX_CMGF_RETRIES) {
+	if (data->retries == MAX_CMGF_RETRIES) {
 		ofono_debug("Unable to enter PDU mode");
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 	}
 
-	g_timeout_add_seconds(1, set_cmgf, modem);
+	g_timeout_add_seconds(1, set_cmgf, sms);
 }
 
 static gboolean set_cmgf(gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 
-	g_at_chat_send(at->parser, "AT+CMGF=0", cmgf_prefix,
-			at_cmgf_set_cb, modem, NULL);
+	g_at_chat_send(data->chat, "AT+CMGF=0", cmgf_prefix,
+			at_cmgf_set_cb, sms, NULL);
 	return FALSE;
 }
 
 static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
 	dump_response("cpms_query_cb", ok, result);
@@ -822,12 +917,12 @@ static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 
 		if (sm_supported[0] && sm_supported[1]) {
 			supported = TRUE;
-			at->sms->store = SM_STORE;
+			data->store = SM_STORE;
 		}
 
 		if (me_supported[0] && me_supported[1]) {
 			supported = TRUE;
-			at->sms->store = ME_STORE;
+			data->store = ME_STORE;
 		}
 
 		/* This seems to be a special case, where the modem will
@@ -835,26 +930,26 @@ static void at_cpms_query_cb(gboolean ok, GAtResult *result,
 		 * mem1
 		 */
 		if (mt_supported[2] && (sm_supported[0] || me_supported[0]))
-			at->sms->incoming = MT_STORE;
+			data->incoming = MT_STORE;
 
 		if (sm_supported[2])
-			at->sms->incoming = SM_STORE;
+			data->incoming = SM_STORE;
 
 		if (me_supported[2])
-			at->sms->incoming = ME_STORE;
+			data->incoming = ME_STORE;
 	}
 out:
 	if (!supported)
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 
-	set_cmgf(modem);
+	set_cmgf(sms);
 }
 
 static void at_cmgf_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
 	dump_response("cmgf_query_cb", ok, result);
@@ -879,17 +974,17 @@ static void at_cmgf_query_cb(gboolean ok, GAtResult *result,
 
 out:
 	if (!supported)
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 
-	g_at_chat_send(at->parser, "AT+CPMS=?", cpms_prefix,
-			at_cpms_query_cb, modem, NULL);
+	g_at_chat_send(data->chat, "AT+CPMS=?", cpms_prefix,
+			at_cpms_query_cb, sms, NULL);
 }
 
 static void at_csms_status_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean supported = FALSE;
 
 	dump_response("csms_status_cb", ok, result);
@@ -915,7 +1010,7 @@ static void at_csms_status_cb(gboolean ok, GAtResult *result,
 			goto out;
 
 		if (service == 1)
-			at->sms->cnma_enabled = TRUE;
+			data->cnma_enabled = TRUE;
 
 		if (mt == 1 && mo == 1)
 			supported = TRUE;
@@ -923,28 +1018,28 @@ static void at_csms_status_cb(gboolean ok, GAtResult *result,
 
 out:
 	if (!supported)
-		return at_sms_not_supported(modem);
+		return at_sms_not_supported(sms);
 
 	/* Now query supported text format */
-	g_at_chat_send(at->parser, "AT+CMGF=?", cmgf_prefix,
-			at_cmgf_query_cb, modem, NULL);
+	g_at_chat_send(data->chat, "AT+CMGF=?", cmgf_prefix,
+			at_cmgf_query_cb, sms, NULL);
 }
 
 static void at_csms_set_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 
-	g_at_chat_send(at->parser, "AT+CSMS?", csms_prefix,
-			at_csms_status_cb, modem, NULL);
+	g_at_chat_send(data->chat, "AT+CSMS?", csms_prefix,
+			at_csms_status_cb, sms, NULL);
 }
 
 static void at_csms_query_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
-	struct ofono_modem *modem = user_data;
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct ofono_sms *sms = user_data;
+	struct sms_data *data = ofono_sms_get_data(sms);
 	gboolean cnma_supported = FALSE;
 	GAtResultIter iter;
 	int status;
@@ -953,12 +1048,7 @@ static void at_csms_query_cb(gboolean ok, GAtResult *result,
 	dump_response("csms_query_cb", ok, result);
 
 	if (!ok)
-		return at_sms_not_supported(modem);
-
-	at->sms = sms_create();
-
-	if (!at->sms)
-		return;
+		return at_sms_not_supported(sms);
 
 	g_at_result_iter_init(&iter, result);
 
@@ -976,27 +1066,53 @@ static void at_csms_query_cb(gboolean ok, GAtResult *result,
 
 out:
 	sprintf(buf, "AT+CSMS=%d", cnma_supported ? 1 : 0);
-	g_at_chat_send(at->parser, buf, csms_prefix,
-			at_csms_set_cb, modem, NULL);
+	g_at_chat_send(data->chat, buf, csms_prefix,
+			at_csms_set_cb, sms, NULL);
 }
 
-void at_sms_init(struct ofono_modem *modem)
+static int at_sms_probe(struct ofono_sms *sms, unsigned int vendor,
+				void *user)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	GAtChat *chat = user;
+	struct sms_data *data;
 
-	g_at_chat_send(at->parser, "AT+CSMS=?", csms_prefix,
-			at_csms_query_cb, modem, NULL);
+	data = g_new0(struct sms_data, 1);
+	data->chat = chat;
+	data->vendor = vendor;
+
+	ofono_sms_set_data(sms, data);
+
+	g_at_chat_send(chat, "AT+CSMS=?", csms_prefix,
+			at_csms_query_cb, sms, NULL);
+
+	return 0;
 }
 
-void at_sms_exit(struct ofono_modem *modem)
+static void at_sms_remove(struct ofono_sms *sms)
 {
-	struct at_data *at = ofono_modem_userdata(modem);
+	struct sms_data *data = ofono_sms_get_data(sms);
 
-	if (!at->sms)
-		return;
+	if (data->cnma_ack_pdu)
+		g_free(data->cnma_ack_pdu);
 
-	sms_destroy(at->sms);
-	at->sms = NULL;
+	g_free(data);
+}
 
-	ofono_sms_manager_unregister(modem);
+static struct ofono_sms_driver driver = {
+	.name = "atmodem",
+	.probe = at_sms_probe,
+	.remove = at_sms_remove,
+	.sca_query	= at_csca_query,
+	.sca_set	= at_csca_set,
+	.submit		= at_cmgs,
+};
+
+void at_sms_init()
+{
+	ofono_sms_driver_register(&driver);
+}
+
+void at_sms_exit()
+{
+	ofono_sms_driver_unregister(&driver);
 }
